from .parameters import Parameters, load_params, save_params
from .utils import sample_many_runs, calc_rre_traj, plot_trajectories

from .cnvm import CNVMParameters, CNVM
from .cntm import CNTMParameters, CNTM



from __future__ import annotations
from typing import Protocol
import networkx as nx
import pickle


class Parameters(Protocol):
    num_opinions: int
    num_agents: int
    network: nx.Graph | None


def save_params(filename: str, params: Parameters):
    """
    Save parameters as pickled file.

    Parameters
    ----------
    filename : str
    params : CNVMParameters
    """
    with open(filename, "wb") as file:
        pickle.dump(params, file)


def load_params(filename: str) -> Parameters:
    """
    Load parameters from pickled file.

    Parameters
    ----------
    filename : str

    Returns
    -------
    CNVMParameters
    """
    with open(filename, "rb") as file:
        return pickle.load(file)



import numpy as np
from numba import njit
from scipy.integrate import solve_ivp
import multiprocessing as mp
from matplotlib.axes import Axes

from .collective_variables import CollectiveVariable
from .parameters import Parameters

from .cnvm.parameters import CNVMParameters
from .cnvm.model import CNVM

from .cntm.parameters import CNTMParameters
from .cntm.model import CNTM


def sample_many_runs(
    params: Parameters,
    initial_states: np.ndarray,
    t_max: float,
    num_timesteps: int,
    num_runs: int,
    n_jobs: int = None,
    collective_variable: CollectiveVariable = None,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Sample multiple runs of the model specified by params.

    Parameters
    ----------
    params : Parameters
        Either CNVM or CNTM Parameters.
        If a NetworkGenerator is used, a new network will be sampled for every run.
    initial_states : np.ndarray
        Array of initial states, shape = (num_initial_states, num_agents).
        Num_runs simulations will be executed for each initial state.
    t_max : float
        End time.
    num_timesteps : int
        Trajectory will be saved at equidistant time points np.linspace(0, t_max, num_timesteps).
    num_runs : int
        Number of samples.
    n_jobs : int, optional
        If "None", no multiprocessing is applied. If "-1", all available CPUs will be used.
    collective_variable : CollectiveVariable, optional
        If collective variable is specified, the projected trajectory will be returned
        instead of the full trajectory.

    Returns
    -------
    t_out, x_out : tuple[np.ndarray, np.ndarray]
        (t_out, x_out), time_out.shape = (num_timesteps,),
        x_out.shape = (num_initial_states, num_runs, num_timesteps, num_agents)
    """
    t_out = np.linspace(0, t_max, num_timesteps)

    # no multiprocessing
    if n_jobs is None or n_jobs == 1:
        x_out = _sample_many_runs_subprocess(
            params, initial_states, t_max, num_timesteps, num_runs, collective_variable
        )
        return t_out, x_out

    # multiprocessing
    if n_jobs == -1:
        n_jobs = mp.cpu_count()

    if num_runs >= initial_states.shape[0]:  # parallelization along runs
        chunks = _split_runs(num_runs, n_jobs)
        processes = [
            (
                params,
                initial_states,
                t_max,
                num_timesteps,
                chunk,
                collective_variable,
            )
            for chunk in chunks
        ]
        concat_axis = 1

    else:  # parallelization along initial states
        chunks = np.array_split(initial_states, n_jobs)
        processes = [
            (
                params,
                chunk,
                t_max,
                num_timesteps,
                num_runs,
                collective_variable,
            )
            for chunk in chunks
        ]
        concat_axis = 0

    with mp.Pool(n_jobs) as pool:
        x_out = pool.starmap(_sample_many_runs_subprocess, processes)
    x_out = np.concatenate(x_out, axis=concat_axis)

    return t_out, x_out


def _sample_many_runs_subprocess(
    params: Parameters,
    initial_states: np.ndarray,
    t_max: float,
    num_timesteps: int,
    num_runs: int,
    collective_variable: CollectiveVariable = None,
) -> np.ndarray:
    t_out = np.linspace(0, t_max, num_timesteps)
    num_initial_states = initial_states.shape[0]

    if isinstance(params, CNVMParameters):
        model_type = CNVM
    elif isinstance(params, CNTMParameters):
        model_type = CNTM
    else:
        raise ValueError("Parameters not valid.")
    model = model_type(params)

    if collective_variable is None:
        x_out = np.zeros(
            (num_initial_states, num_runs, num_timesteps, model.params.num_agents)
        )
    else:
        x_out = np.zeros(
            (num_initial_states, num_runs, num_timesteps, collective_variable.dimension)
        )

    for j in range(num_initial_states):
        for i in range(num_runs):
            t, x = model.simulate(
                t_max, len_output=4 * num_timesteps, x_init=initial_states[j]
            )
            t_ind = argmatch(t_out, t)
            if collective_variable is None:
                x_out[j, i, :, :] = x[t_ind, :]
            else:
                x_out[j, i, :, :] = collective_variable(x[t_ind, :])
    return x_out


def _split_runs(num_runs: int, num_chunks: int) -> np.ndarray:
    """
    Split num_runs into num_chunks approximately equal chunks.
    """
    chunks = np.ones(num_chunks, dtype=int) * (num_runs // num_chunks)
    chunks[: (num_runs % num_chunks)] += 1
    return chunks


def calc_rre_traj(
    params: CNVMParameters, c_0: np.ndarray, t_max: float, t_eval=None
) -> tuple[np.ndarray, np.ndarray]:
    """
    Solve the RRE given by parameters, starting from c_0, up to time t_max.

    Solves the ODE using scipy's "solve_ivp".

    Parameters
    ----------
    params : CNVMParameters
    c_0 : np.ndarray
        Initial state, shape=(num_opinions,)
    t_max : float
        End time.
    t_eval : np.ndarray, optional
        Time points, at which the solution should be evaluated.

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        1. timepoints, shape=(?,).
        2. c, shape=(?, num_opinions).
    """

    def rhs(t, c):
        out = np.zeros_like(c)
        for m in range(params.num_opinions):
            for n in range(params.num_opinions):
                if n == m:
                    continue

                state_change = np.zeros_like(c)
                state_change[m] = -1
                state_change[n] = 1

                prop = c[m] * (
                    params.r_imit * params.prob_imit[m, n] * c[n]
                    + params.r_noise * params.prob_noise[m, n] / params.num_opinions
                )

                out += prop * state_change
        return out

    sol = solve_ivp(rhs, (0, t_max), c_0, rtol=1e-8, atol=1e-8, t_eval=t_eval)
    return sol.t, sol.y.T


def plot_trajectories(
    params: Parameters,
    t: float,
    cv: CollectiveVariable,
    ax: Axes,
    num_initial_states: int = 3,
    samples_per_state: int = 5,
) -> Axes:
    """
    Plot some trajectories with uniformly random initial states
    in the provided Axes object.

    The purpose of this function is to quickly inspect how trajectories
    look like for the given set of Parameters.

    Parameters
    ----------
    params : Parameters
    t : float
    cv : CollectiveVariable
    ax : Axes
    num_initial_states : int, optional
    samples_per_state : int, optional

    Returns
    -------
    Axes
    """
    initial_states = np.random.randint(
        0, params.num_opinions, size=(num_initial_states, params.num_agents)
    )

    t, c = sample_many_runs(
        params,
        initial_states,
        t,
        1000,
        samples_per_state,
        collective_variable=cv,
    )

    colors = ["k", "b", "g", "c", "r", "y"]
    linestyles = ["-", "--", "-."]

    for i in range(num_initial_states):
        this_linestyle = linestyles[i % len(linestyles)]
        for j in range(samples_per_state):
            this_color = colors[j % len(colors)]
            ax.plot(t, c[i, j, :, 0], linestyle=this_linestyle, color=this_color)

    ax.set_xlabel("$t$")
    ax.set_ylabel("$c_0$")
    ax.grid()

    return ax


@njit
def argmatch(x_ref, x):
    """
    Find indices such that |x[indices] - x_ref| = min!

    Parameters
    ----------
    x_ref : np.ndarray
        1D, sorted
    x : np.ndarray
        1D, sorted
    Returns
    -------
    np.ndarray
    """
    size = np.shape(x_ref)[0]
    out = np.zeros(size, dtype=np.int64)
    ref_ind = 0
    ind = 0
    while x_ref[ref_ind] < x[ind]:
        ref_ind += 1

    while ref_ind < size and ind < x.shape[0] - 1:
        if x[ind] <= x_ref[ref_ind] <= x[ind + 1]:
            if np.abs(x[ind] - x_ref[ref_ind]) < np.abs(
                x[ind + 1] - x_ref[ref_ind]
            ):  # smaller is nearer
                out[ref_ind] = ind
            else:  # bigger is nearer
                out[ref_ind] = ind + 1
            ref_ind += 1
        else:
            ind += 1

    while ref_ind < size:
        out[ref_ind] = x.shape[0] - 1
        ref_ind += 1

    return out



from __future__ import annotations
from dataclasses import dataclass, field
import networkx as nx
import numpy as np

from ..network_generator import NetworkGenerator


@dataclass()
class CNVMParameters:
    """
    Container for the parameters of the Continuous-time Noisy Voter Model (CNVM).

    A node i transitions from its current opinion m to a different opinion n at rate
    r[m,n] * d(i,n) / (d(i)^alpha) + r_tilde[m,n],
    where d(i,n) is the count of opinion n in the neighborhood of agent i, and d(i) the degree of node i.

    Either a network has to specified, or a NetworkGenerator,
    or num_agents, in which case a complete network is used.
    If multiple are given, NetworkGenerator overrules network, and network overrules num_agents.

    The rate parameters r and r_tilde can be given as numpy arrays of shape (num_opinions, num_opinions),
    or as floats, in which case all rates are set to this value.

    (Internally, the CNVM uses an equivalent set of rate parameters: r_imit, r_noise, prob_imit and prob_noise.
    A node i transitions from its current opinion m to a different opinion n at rate
    r_imit * d(i,n) / (d(i)^alpha) * prob_imit[m, n] + r_noise * (1/num_opinions) * prob_noise[m, n].
    These parameters can be provided instead of the usual r and r_tilde.)
    """

    num_opinions: int
    num_agents: int = None
    network: nx.Graph | None = field(default=None, repr=False)
    network_generator: NetworkGenerator | None = None
    alpha: float = 1

    # rate parameters in style 1
    r: float | np.ndarray = None
    r_tilde: float | np.ndarray = None

    # rate parameters in style 2
    r_imit: float = None
    r_noise: float = None
    prob_imit: float | np.ndarray = 1
    prob_noise: float | np.ndarray = 1

    def __post_init__(self):
        # rates
        self._tidy_up_rates()

        # networks
        if self.network_generator is not None:
            self.num_agents = self.network_generator.num_agents
            self.network = None
        elif self.network is not None:
            self.num_agents = len(self.network.nodes)
        elif self.num_agents is None:
            raise ValueError(
                "Either a network or a NetworkGenerator or num_agents has to be specified."
            )

    def _tidy_up_rates(self):
        """
        Check if the given rate parameters are valid and set up the other rate parameters accordingly.
        """
        one_mat = np.ones((self.num_opinions, self.num_opinions))
        if self.r is not None and self.r_tilde is not None:  # style 1
            if isinstance(self.r, (float, int)):
                self.r = self.r * one_mat
            if isinstance(self.r_tilde, (float, int)):
                self.r_tilde = self.r_tilde * one_mat
            np.fill_diagonal(self.r, 0)
            np.fill_diagonal(self.r_tilde, 0)

            if np.min(self.r) < 0 or np.min(self.r_tilde) < 0:
                raise ValueError("Rates have to be non-negative.")

            (
                self.r_imit,
                self.r_noise,
                self.prob_imit,
                self.prob_noise,
            ) = convert_rate_to_cnvm(self.r, self.r_tilde)

        elif self.r_imit is not None and self.r_noise is not None:  # style 2
            if isinstance(self.prob_imit, (float, int)):
                self.prob_imit = self.prob_imit * one_mat
            if isinstance(self.prob_noise, (float, int)):
                self.prob_noise = self.prob_noise * one_mat
            np.fill_diagonal(self.prob_imit, 0)
            np.fill_diagonal(self.prob_noise, 0)

            if np.min(self.r_imit) < 0 or np.min(self.r_noise) < 0:
                raise ValueError("Rates have to be non-negative.")
            if np.min(self.prob_imit) < 0 or np.max(self.prob_imit) > 1:
                raise ValueError("Probabilities have to be between 0 and 1.")
            if np.min(self.prob_noise) < 0 or np.max(self.prob_noise) > 1:
                raise ValueError("Probabilities have to be between 0 and 1.")

            self.r = self.r_imit * self.prob_imit
            self.r_tilde = self.r_noise * self.prob_noise / self.num_opinions

        else:
            raise ValueError("Rate parameters have to be provided.")

    def change_rates(
        self, r: float | np.ndarray = None, r_tilde: float | np.ndarray = None
    ):
        """
        Change one or both rate parameters.

        If only one argument is given, the other rate parameter stays the same.

        Parameters
        ----------
        r : float | np.ndarray, optional
        r_tilde : float | np.ndarray
        """
        if r is not None:
            self.r = r
        if r_tilde is not None:
            self.r_tilde = r_tilde
        self._tidy_up_rates()

    def get_network(self) -> nx.Graph:
        """
        If self.network exists, returns the network.
        Else, if a NetworkGenerator was given, returns a network generated by it.
        Else, returns the complete graph.

        Returns
        -------
        nx.Graph
        """
        if self.network is not None:
            return self.network
        elif self.network_generator is not None:
            return self.network_generator()
        else:
            return nx.complete_graph(self.num_agents)

    def set_network(self, network: nx.Graph) -> None:
        """
        Set new network.

        Only works if no NetworkGenerator was given.

        Parameters
        ----------
        network : nx.Graph
        """
        if self.network_generator is not None:
            raise ValueError("Cannot set network when there is a NetworkGenerator.")
        self.network = network
        self.num_agents = len(network.nodes)

    def update_network_by_generator(self) -> None:
        """
        Update self.network via the NetworkGenerator.

        Only works if a NetworkGenerator was given.
        """
        if self.network_generator is None:
            raise ValueError("No NetworkGenerator was given.")
        self.network = self.network_generator()


def save_params_as_txt_file(filename: str, params: CNVMParameters):
    """
    Save parameters as a readable .txt file.

    Parameters
    ----------
    filename : str
    params : CNVMParameters
    """
    if filename[-4:] != ".txt":
        this_filename = filename + ".txt"
    else:
        this_filename = filename

    with open(this_filename, "w") as f:
        f.write(f"num_agents = {params.num_agents}\n")
        f.write(f"num_opinions = {params.num_opinions}\n")
        f.write(f"alpha = {params.alpha}\n")
        f.write(f"r_imit = {params.r_imit}\n")
        f.write(f"r_noise = {params.r_noise}\n\n")
        f.write(f"prob_imit =\n {params.prob_imit}\n\n")
        f.write(f"prob_noise =\n {params.prob_noise}\n\n")

        if params.network_generator is not None:
            f.write(f"network_generator = {params.network_generator}\n")
        elif params.network is not None:
            f.write(f"network = {params.network}\n")
        else:
            f.write(f"network = fully connected\n")


def convert_rate_to_cnvm(
    r: np.ndarray, r_tilde: np.ndarray
) -> tuple[float, float, np.ndarray, np.ndarray]:
    """
    Convert the rates r and r_tilde to the parameters used in the CNVM, i.e., r_imit, r_noise, prob_imit, prob_noise.

    The rates r and r_tilde are defined as:
    An agent i transitions from his current opinion m to a different opinion n at rate
    r[m, n] * d(i,n) / (d(i)^alpha) + r_tilde[m, n]
    where d(i,n) is the count of opinion n in the neighborhood of agent i, and d(i) the degree of node i.

    Parameters
    ----------
    r : np.ndarray
        shape = (num_opinions, num_opinions)
    r_tilde : np.ndarray
        shape = (num_opinions, num_opinions)

    Returns
    -------
    tuple[float, float, np.ndarray, np.ndarray]
        r_imit, r_noise, prob_imit, prob_noise
    """
    num_opinions = r.shape[0]

    # r[m,n] = r_imit * prob_imit[m,n]
    this_r = np.copy(r)
    np.fill_diagonal(this_r, 0)
    r_imit = np.max(this_r)
    if r_imit > 0:
        prob_imit = this_r / r_imit
    else:
        prob_imit = np.zeros((num_opinions, num_opinions))

    # r_tilde[m, n] = r_noise / num_opinions * prob_noise[m, n]
    this_r_tilde = np.copy(r_tilde)
    np.fill_diagonal(this_r_tilde, 0)
    r_noise = np.max(this_r_tilde) * num_opinions
    if r_noise > 0:
        prob_noise = this_r_tilde * num_opinions / r_noise
    else:
        prob_noise = np.zeros((num_opinions, num_opinions))

    return r_imit, r_noise, prob_imit, prob_noise


def convert_rate_from_cnvm(params: CNVMParameters) -> tuple[np.ndarray, np.ndarray]:
    """
    Convert the rates used in the CNVM to r and r_tilde.

    The rates r and r_tilde are defined as:
    An agent i transitions from his current opinion m to a different opinion n at rate
    r[m, n] * d(i,n) / (d(i)^alpha) + r_tilde[m, n]
    where d(i,n) is the count of opinion n in the neighborhood of agent i, and d(i) the degree of node i.

    Parameters
    ----------
    params : CNVMParameters

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        r, r_tilde
    """
    r = params.r_imit * params.prob_imit
    r_tilde = params.r_noise * params.prob_noise / params.num_opinions
    return r, r_tilde



from typing import Protocol, Union

import networkx as nx
import numpy as np
from numba import njit

from .cnvm.parameters import CNVMParameters


class CollectiveVariable(Protocol):
    dimension: int

    def __call__(self, x_traj: np.ndarray) -> np.ndarray:
        """
        Parameters
        ----------
        x_traj : np.ndarray
            trajectory of CNVM, shape = (?, num_agents).

        Returns
        -------
        np.ndarray
            trajectory projected down via the collective variable, shape = (?, self.dimension)
        """


class OpinionShares:
    def __init__(
        self,
        num_opinions,
        normalize: bool = False,
        weights: np.ndarray = None,
        idx_to_return: Union[int, np.ndarray] = None,
    ):
        """
        Calculate the opinion counts/ percentages, i.e., how often each opinion is present in x.

        Parameters
        ----------
        num_opinions : int
        normalize : bool, optional
            If true return percentages, else counts.
        weights : np.ndarray, optional
            Weight for each agent's opinion, shape=(num_agents,). Default: Each agent has weight 1.
            Negative weights are allowed.
        idx_to_return : Union[int, np.ndarray], optional
            Shares of which opinions to return. Default: all opinions.
            Example: idx_to_return=0 means that only the count of opinion 0 is returned.
        """
        self.num_opinions = num_opinions
        self.normalize = normalize
        self.weights = weights

        if idx_to_return is None:
            self.idx_to_return = np.arange(num_opinions)
        elif isinstance(idx_to_return, int):
            self.idx_to_return = np.array([idx_to_return])
        else:
            self.idx_to_return = idx_to_return

        self.dimension = len(self.idx_to_return)

    def __call__(self, x_traj: np.ndarray) -> np.ndarray:
        """
        Parameters
        ----------
        x_traj : np.ndarray
            trajectory of CNVM, shape = (?, num_agents).

        Returns
        -------
        np.ndarray
            trajectory projected down via the collective variable, shape = (?, self.dimension)
        """
        num_agents = x_traj.shape[1]
        x_agg = _opinion_shares_numba(
            x_traj.astype(int), self.num_opinions, self.weights
        )
        x_agg = x_agg[:, self.idx_to_return]

        if self.normalize:
            if self.weights is None:
                x_agg /= num_agents
            else:
                x_agg /= np.sum(np.abs(self.weights))
        return x_agg


class OpinionSharesByDegree:
    def __init__(
        self,
        num_opinions: int,
        network: nx.Graph,
        normalize: bool = False,
        idx_to_return: Union[int, np.ndarray] = None,
    ):
        """
        Calculate the count of each opinion by degree.

        The output has dimension idx_to_return * number of different degrees.
        For example, the first idx_to_return entries will represent the counts for nodes with the smallest degree.

        Parameters
        ----------
        num_opinions : int
        network : nx.Graph
        normalize : bool, optional
            If true return percentages, else counts.
            The normalization is done within each group of nodes with the same degree.
        idx_to_return : Union[int, np.ndarray], optional
            Shares of which opinions to return. Default: all opinions.
        """
        self.degrees_of_nodes = np.array([d for _, d in network.degree()])
        self.degrees = np.sort(np.unique(self.degrees_of_nodes))
        self.num_opinions = num_opinions
        self.normalize = normalize

        if idx_to_return is None:
            self.idx_to_return = np.arange(num_opinions)
        elif isinstance(idx_to_return, int):
            self.idx_to_return = np.array([idx_to_return])
        else:
            self.idx_to_return = idx_to_return

        self.dimension = len(self.idx_to_return) * self.degrees.shape[0]

    def __call__(self, x_traj: np.ndarray) -> np.ndarray:
        """
        Parameters
        ----------
        x_traj : np.ndarray
            trajectory of CNVM, shape = (?, num_agents).

        Returns
        -------
        np.ndarray
            trajectory projected down via the collective variable, shape = (?, self.dimension)
        """
        cv = np.zeros((x_traj.shape[0], self.dimension))
        num_agents = x_traj.shape[1]
        x_traj_int = x_traj.astype(int)

        for i, deg in enumerate(self.degrees):
            weights = np.zeros(num_agents)
            weights[np.nonzero(self.degrees_of_nodes == deg)] = 1
            x_agg = _opinion_shares_numba(x_traj_int, self.num_opinions, weights)
            x_agg = x_agg[:, self.idx_to_return]
            if self.normalize:
                x_agg /= np.sum(weights)
            cv[
                :, i * len(self.idx_to_return) : (i + 1) * len(self.idx_to_return)
            ] = np.copy(x_agg)

        return cv


class CompositeCollectiveVariable:
    def __init__(self, collective_variables: list[CollectiveVariable]):
        """
        Concatenate multiple collective variables.

        Typical use-case: CV1 measures the share of opinion 1 in one part of the network,
        CV2 in a different part of the network (both built via OpinionShares class with weights).
        CompositeCollectiveVariable([CV1, CV2]) concatenates the output of the two.

        Parameters
        ----------
        collective_variables : list
        """
        self.collective_variables = collective_variables
        self.dimension = sum([cv.dimension for cv in collective_variables])

    def __call__(self, x_traj: np.ndarray) -> np.ndarray:
        """
        Parameters
        ----------
        x_traj : np.ndarray
            trajectory of CNVM, shape = (?, num_agents).

        Returns
        -------
        np.ndarray
            trajectory projected down via the collective variable, shape = (?, self.dimension)
        """
        return np.concatenate([cv(x_traj) for cv in self.collective_variables], axis=1)


class Interfaces:
    def __init__(self, network: nx.Graph, normalize: bool = False):
        """
        Count the number of interfaces between opinion 0 and 1.

        Can not be used when there are more than these two opinions.

        Parameters
        ----------
        network : nx.Graph
        normalize : bool, optional
            Normalize by dividing by the number of edges in the network.
        """
        self.dimension = 1
        self.normalize = normalize
        self.network = network

    def __call__(self, x_traj: np.ndarray) -> np.ndarray:
        """
        Parameters
        ----------
        x_traj : np.ndarray
            trajectory of CNVM, shape = (?, num_agents).

        Returns
        -------
        np.ndarray
            trajectory projected down via the collective variable, shape = (?, self.dimension)
        """
        if np.max(x_traj) > 1:
            raise ValueError("Interfaces can only be used for 2 opinions.")
        out = np.zeros((x_traj.shape[0], 1))

        for i in range(x_traj.shape[0]):
            for u, v in self.network.edges:
                if x_traj[i, u] != x_traj[i, v]:
                    out[i, 0] += 1

        if self.normalize:
            out /= self.network.number_of_edges()

        return out


class Propensities:
    def __init__(self, params: CNVMParameters):
        """
        The propensities are defined as cumulative transition rates in the system.

        Only implemented for 2 opinions, 0 and 1.
        Output 2-dimensional, (prop_01, prop_10).

        The propensity prop_mn is defined as
        sum_i^N ( r[m, n] * d(i,n) / (d(i)^alpha) + r_tilde[m, n] ).

        Parameters
        ----------
        params : CNVMParameters
        """
        self.dimension = 2
        self.params = params

    def __call__(self, x_traj: np.ndarray) -> np.ndarray:
        out = np.zeros((x_traj.shape[0], 2))
        for j in range(self.params.num_agents):
            if self.params.network is not None:
                neighbors = list(self.params.network.neighbors(j))
                degrees = [d for _, d in self.params.network.degree()]
            else:  # complete
                neighbors = list(range(0, j)) + list(
                    range(j + 1, self.params.num_agents)
                )
                degrees = (self.params.num_agents - 1) * np.ones(self.params.num_agents)

            for i in range(x_traj.shape[0]):
                if x_traj[i, j] == 0:
                    m, n = 0, 1
                else:
                    m, n = 1, 0
                count_opinion_n = np.sum(x_traj[i, neighbors] == n)

                prop_mn = (
                    self.params.r[m, n]
                    * count_opinion_n
                    / (degrees[i] ** self.params.alpha)
                )
                prop_mn += self.params.r_tilde[m, n]
                out[i, m] += prop_mn

        return out


@njit
def _opinion_shares_numba(x_traj, num_opinions, weights):
    x_agg = np.zeros((x_traj.shape[0], num_opinions))
    for i in range(x_traj.shape[0]):
        x_agg[i, :] = np.bincount(x_traj[i, :], weights=weights, minlength=num_opinions)
    return x_agg



from dataclasses import dataclass
import networkx as nx


@dataclass()
class CNTMParameters:
    """
    Container for the parameters of the Threshold Model.

    At the rate r, each node evaluates to change their opinion from its current
    opinion m=0,1 to the other opinion n=1-m. It changes the opinion if the
    percentage of neighbors of opinion n exceeds the threshold_mn.

    Additionally, each node changes its state randomly at rate r_tilde (noise).
    """

    network: nx.Graph
    r: float
    r_tilde: float
    threshold_01: float
    threshold_10: float

    def __post_init__(self):
        self.num_agents = self.network.number_of_nodes()
        self.num_opinions = 2



import numpy as np
from numpy.random import Generator, default_rng
from numba import njit
from numba.typed import List
from .parameters import CNTMParameters


class CNTM:
    def __init__(self, params: CNTMParameters):
        """
        Continuous-time Noisy Threshold Model.

        Parameters
        ----------
        params : CNTMParameters
        """
        self.params = params

        self.neighbor_list = (
            List()
        )  # self.neighbor_list[i] = array of neighbors of node i
        for i in range(self.params.num_agents):
            self.neighbor_list.append(
                np.array(list(self.params.network.neighbors(i)), dtype=int)
            )

        self.noise_prob = self.params.r_tilde / (self.params.r + self.params.r_tilde)
        self.next_event_rate = 1 / (
            self.params.num_agents * (self.params.r + self.params.r_tilde)
        )

    def simulate(
        self,
        t_max: float,
        x_init: np.ndarray,
        len_output: int = None,
        rng: Generator = default_rng(),
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Simulate the model from t=0 to t=t_max.

        Parameters
        ----------
        t_max : float
        x_init : np.ndarray
            shape=(num_agents,)
        len_output : int, optional
            number of snapshots to output, as equidistantly spaced as possible between 0 and t_max
        rng : Generator, optional
            random number generator

        Returns
        -------
        tuple[np.ndarray]
            t_traj (shape=(?,)), x_traj (shape=(?,num_agents))
        """
        x = np.copy(x_init).astype(float)

        t_delta = 0 if len_output is None else t_max / (len_output - 1)

        t_traj, x_traj = _simulate_numba(
            x,
            t_delta,
            self.next_event_rate,
            self.noise_prob,
            t_max,
            self.params.num_agents,
            self.params.threshold_01,
            self.params.threshold_10,
            self.neighbor_list,
            rng,
        )

        return np.array(t_traj), np.array(x_traj)


@njit
def _simulate_numba(
    x,
    t_delta,
    next_event_rate,
    noise_prob,
    t_max,
    num_agents,
    threshold_01,
    threshold_10,
    neighbor_list,
    rng,
):
    x_traj = [np.copy(x)]
    t = 0
    t_traj = [0]

    t_store = t_delta
    while t < t_max:
        t += rng.exponential(next_event_rate)  # time of next event
        agent = rng.integers(0, num_agents)  # agent of next event

        if rng.random() < noise_prob:  # noise
            x[agent] = rng.integers(0, 2)
        else:
            neighbors = neighbor_list[agent]
            if len(neighbors) > 0:
                other_opinion = 1 - x[agent]
                share_other_opinion = 0
                for j in neighbors:
                    if x[j] == other_opinion:
                        share_other_opinion += 1
                share_other_opinion /= len(neighbors)

                threshold = threshold_10 if x[agent] == 1 else threshold_01
                if share_other_opinion >= threshold:
                    x[agent] = other_opinion

        if t >= t_store:
            t_store += t_delta
            x_traj.append(x.copy())
            t_traj.append(t)

    return t_traj, x_traj



import numba
import numpy as np
from .parameters import CNVMParameters


def sample_cle(
    params: CNVMParameters,
    initial_state: np.ndarray,
    max_time: float,
    num_time_steps: int,
    num_samples: int,
    saving_offset: int = 1,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Sample Chemical Langevin Equation (CLE) approximation for the CNVM.

    The Euler-Maruyama method is used to integrate the SDE.

    Parameters
    ----------
    params : CNVMParameters
    initial_state : np.ndarray
        Shape = (num_opinions,)
    max_time : float
    num_time_steps : int
        The step size of the integration is max_time / num_time_steps.
    num_samples : int
    saving_offset : int, optional
        Only return every saving_offset-th state to save memory.

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        (t, c), t.shape=(num_time_steps + 1), c.shape = (num_samples, num_time_steps + 1, num_opinions).
        (If saving_offset > 1, the number of time steps will be smaller.)
    """
    return _numba_sample_cle(
        initial_state,
        max_time,
        num_time_steps,
        params.num_agents,
        params.r,
        params.r_tilde,
        num_samples,
        saving_offset,
    )


@numba.njit(parallel=True)
def _numba_sample_cle(
    initial_state: np.ndarray,
    max_time: float,
    num_time_steps: int,
    num_agents: int,
    r: np.ndarray,
    r_tilde: np.ndarray,
    num_samples: int,
    saving_offset: int,
) -> tuple[np.ndarray, np.ndarray]:
    dim = initial_state.shape[0]
    t = np.linspace(0, max_time, num_time_steps + 1)
    t = t[::saving_offset]
    x_out = np.zeros((num_samples, t.shape[0], dim))

    for i in numba.prange(num_samples):
        x = _numba_euler_maruyama(
            initial_state, max_time, num_time_steps, num_agents, r, r_tilde
        )
        x_out[i] = x[::saving_offset, :]

    return t, x_out


@numba.njit()
def _numba_euler_maruyama(
    initial_state: np.ndarray,
    max_time: float,
    num_time_steps: int,
    num_agents: int,
    r: np.ndarray,
    r_tilde: np.ndarray,
) -> np.ndarray:
    dim = initial_state.shape[0]
    x = np.zeros((num_time_steps + 1, dim))

    x[0] = np.copy(initial_state)
    delta_t = max_time / num_time_steps
    dim_diffusion = dim**2 - dim
    wiener_increments = np.random.normal(
        0, delta_t**0.5, (num_time_steps, dim_diffusion)
    )

    for i in range(num_time_steps):
        drift, diffusion = _drift_and_diffusion(x[i], r, r_tilde, num_agents)
        x[i + 1] = x[i] + drift * delta_t + diffusion @ wiener_increments[i]
        x[i + 1] = np.clip(x[i + 1], 0, 1)

    return x


@numba.njit()
def _drift_and_diffusion(c, r, r_tilde, num_agents):
    num_o = c.shape[0]
    drift = np.zeros(num_o)
    diffusion = np.zeros((num_o, num_o**2 - num_o))

    i = 0
    for m in range(num_o):
        for n in range(num_o):
            if n == m:
                continue

            state_change = np.zeros(num_o)
            state_change[m] = -1
            state_change[n] = 1
            prop = c[m] * (r[m, n] * c[n] + r_tilde[m, n])
            drift += prop * state_change
            diffusion[:, i] = (prop / num_agents) ** 0.5 * state_change
            i += 1

    return drift, diffusion



from __future__ import annotations
import numpy as np
from numpy.random import Generator, default_rng
from numba import njit
from numba.typed import List

from .parameters import CNVMParameters


class CNVM:
    def __init__(self, params: CNVMParameters):
        """
        Continuous-time Noisy Voter Model.

        Parameters
        ----------
        params : CNVMParameters
        """
        self.params = params
        # self.neighbor_list[i] = array of neighbors of node i
        self.neighbor_list = None
        self.degree_alpha = None  # array containing d(i)^(1 - alpha)

        self._calculate_degree_alpha()
        self._calculate_neighbor_list()

    def _calculate_neighbor_list(self):
        """
        Calculate and set self.neighbor_list.
        """
        self.neighbor_list = List()
        if self.params.network is not None:  # not needed for complete network
            for i in range(self.params.num_agents):
                self.neighbor_list.append(
                    np.array(list(self.params.network.neighbors(i)), dtype=int)
                )

    def _calculate_degree_alpha(self):
        """
        Calculate and set self.degree_alpha.
        """
        if self.params.network is not None:
            degrees = np.array([d for _, d in self.params.network.degree()])
            if np.min(degrees) < 1:
                raise ValueError("Isolated vertices in the network are not allowed.")
            self.degree_alpha = degrees ** (1 - self.params.alpha)
        else:  # fully connected
            self.degree_alpha = np.ones(self.params.num_agents) * (
                self.params.num_agents - 1
            ) ** (1 - self.params.alpha)

    def update_network(self):
        """
        Update network from NetworkGenerator in params.
        """
        self.params.update_network_by_generator()
        self._calculate_degree_alpha()
        self._calculate_neighbor_list()

    def update_rates(
        self, r: float | np.ndarray = None, r_tilde: float | np.ndarray = None
    ):
        """
        Update one or both rate parameters.

        If only one argument is given, the other rate parameter stays the same.

        Parameters
        ----------
        r : float | np.ndarray, optional
        r_tilde : float | np.ndarray
        """
        self.params.change_rates(r, r_tilde)

    def simulate(
        self,
        t_max: float,
        x_init: np.ndarray = None,
        len_output: int = None,
        rng: Generator = default_rng(),
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Simulate the model from t=0 to t=t_max.

        Parameters
        ----------
        t_max : float
        x_init : np.ndarray, optional
            Initial state, shape=(num_agents,). If no x_init is given, a random one is generated.
        len_output : int, optional
            Number of snapshots to output, as equidistantly spaced as possible between 0 and t_max.
            Needs to be at least 2 (for initial value and final value).
        rng : Generator, optional
            random number generator

        Returns
        -------
        tuple[np.ndarray, np.ndarray]
            t_traj (shape=(?,)), x_traj (shape=(?,num_agents))
        """
        if self.params.network_generator is not None:
            self.update_network()

        if x_init is None:
            x_init = rng.choice(
                np.arange(self.params.num_opinions), size=self.params.num_agents
            )
        x = np.copy(x_init).astype(int)

        t_delta = 0 if len_output is None else t_max / (len_output - 1)

        # for complete networks, we have a separate implementation
        if self.params.network is None:
            t_traj, x_traj = _numba_simulate_const_complete(
                x,
                t_delta,
                t_max,
                self.params.num_opinions,
                self.params.r_imit,
                self.params.r_noise,
                self.params.prob_imit,
                self.params.prob_noise,
                self.degree_alpha,
                rng,
            )

        else:
            # decide which version is likely faster, log-update or constant-update
            val = np.mean(self.degree_alpha / np.max(self.degree_alpha))
            if 1 / val < np.log2(self.params.num_agents):
                sim_function = _numba_simulate_const
            else:
                sim_function = _numba_simulate_log

            t_traj, x_traj = sim_function(
                x,
                t_delta,
                t_max,
                self.params.num_opinions,
                self.neighbor_list,
                self.params.r_imit,
                self.params.r_noise,
                self.params.prob_imit,
                self.params.prob_noise,
                self.degree_alpha,
                rng,
            )

        return np.array(t_traj), np.array(x_traj, dtype=int)

    def simulate_log_complexity(
        self,
        t_max: float,
        x_init: np.ndarray = None,
        len_output: int = None,
        rng: Generator = default_rng(),
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Simulate the model from t=0 to t=t_max using the log-complexity update.

        Depending on the model parameters, the log-complexity update can be faster
        than the constant-complexity update because the constant-complexity update
        more often performs iterations that do not result in a transition.


        Parameters
        ----------
        t_max : float
        x_init : np.ndarray, optional
            Initial state, shape=(num_agents,). If no x_init is given, a random one is generated.
        len_output : int, optional
            Number of snapshots to output, as equidistantly spaced as possible between 0 and t_max.
            Needs to be at least 2 (for initial value and final value).
        rng : Generator, optional
            random number generator

        Returns
        -------
        tuple[np.ndarray, np.ndarray]
            t_traj (shape=(?,)), x_traj (shape=(?,num_agents))
        """
        if self.params.network_generator is not None:
            self.update_network()

        if x_init is None:
            x_init = rng.choice(
                np.arange(self.params.num_opinions), size=self.params.num_agents
            )
        x = np.copy(x_init).astype(int)

        t_delta = 0 if len_output is None else t_max / (len_output - 1)

        t_traj, x_traj = _numba_simulate_log(
            x,
            t_delta,
            t_max,
            self.params.num_opinions,
            self.neighbor_list,
            self.params.r_imit,
            self.params.r_noise,
            self.params.prob_imit,
            self.params.prob_noise,
            self.degree_alpha,
            rng,
        )

        return np.array(t_traj), np.array(x_traj, dtype=int)

    def simulate_const_complexity(
        self,
        t_max: float,
        x_init: np.ndarray = None,
        len_output: int = None,
        rng: Generator = default_rng(),
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Simulate the model from t=0 to t=t_max using the constant-complexity update.

        Depending on the model parameters, the log-complexity update can be faster
        than the constant-complexity update because the constant-complexity update
        more often performs iterations that do not result in a transition.


        Parameters
        ----------
        t_max : float
        x_init : np.ndarray, optional
            Initial state, shape=(num_agents,). If no x_init is given, a random one is generated.
        len_output : int, optional
            Number of snapshots to output, as equidistantly spaced as possible between 0 and t_max.
            Needs to be at least 2 (for initial value and final value).
        rng : Generator, optional
            random number generator

        Returns
        -------
        tuple[np.ndarray, np.ndarray]
            t_traj (shape=(?,)), x_traj (shape=(?,num_agents))
        """
        if self.params.network_generator is not None:
            self.update_network()

        if x_init is None:
            x_init = rng.choice(
                np.arange(self.params.num_opinions), size=self.params.num_agents
            )
        x = np.copy(x_init).astype(int)

        t_delta = 0 if len_output is None else t_max / (len_output - 1)

        t_traj, x_traj = _numba_simulate_const(
            x,
            t_delta,
            t_max,
            self.params.num_opinions,
            self.neighbor_list,
            self.params.r_imit,
            self.params.r_noise,
            self.params.prob_imit,
            self.params.prob_noise,
            self.degree_alpha,
            rng,
        )

        return np.array(t_traj), np.array(x_traj, dtype=int)


@njit()
def _rand_index_numba(prob_cum_sum: np.ndarray, rng: Generator) -> int:
    """
    Sample random index 0 <= i < len(prob_cumsum) according to probability distribution.

    Parameters
    ----------
    prob_cum_sum : np.ndarray
        1D array containing the cumulative probabilities, i.e.,
        the first entry is the probability of choosing index 0,
        the second entry the probability of choosing index 0 or 1, and so on.
        The last entry is 1.
    rng : Generator
        random number generator

    Returns
    -------
    int
    """
    return np.searchsorted(prob_cum_sum, rng.random(), side="right")


@njit()
def _numba_simulate_log(
    x: np.ndarray,
    t_delta: float,
    t_max: float,
    num_opinions: int,
    neighbor_list: list,
    r_imit: float,
    r_noise: float,
    prob_imit: np.ndarray,
    prob_noise: np.ndarray,
    degree_alpha: np.ndarray,
    rng: Generator,
):
    """
    CNVM simulation with log-complexity update.
    """
    # pre-calculate some values
    num_agents = x.shape[0]
    next_event_rate = 1 / (r_imit * np.sum(degree_alpha) + r_noise * num_agents)
    noise_probability = r_noise * num_agents * next_event_rate
    prob_cum_sum = np.cumsum(degree_alpha / np.sum(degree_alpha))

    # initialize
    x_traj = [np.copy(x)]
    t = 0
    t_traj = [0]

    # simulation loop
    t_store = t_delta
    while t < t_max:
        t += rng.exponential(next_event_rate)  # time of next event
        noise = True if rng.random() < noise_probability else False

        if noise:
            agent = rng.integers(0, num_agents)  # agent of next event
            new_opinion = rng.integers(0, num_opinions)
            if rng.random() < prob_noise[x[agent], new_opinion]:
                x[agent] = new_opinion
        else:
            agent = _rand_index_numba(prob_cum_sum, rng)
            neighbors = neighbor_list[agent]
            rand_neighbor = neighbors[rng.integers(0, len(neighbors))]
            new_opinion = x[rand_neighbor]
            if rng.random() < prob_imit[x[agent], new_opinion]:
                x[agent] = new_opinion

        if t >= t_store:
            t_store += t_delta
            x_traj.append(x.copy())
            t_traj.append(t)

    return t_traj, x_traj


@njit()
def _numba_simulate_const(
    x: np.ndarray,
    t_delta: float,
    t_max: float,
    num_opinions: int,
    neighbor_list: list,
    r_imit: float,
    r_noise: float,
    prob_imit: np.ndarray,
    prob_noise: np.ndarray,
    degree_alpha: np.ndarray,
    rng: Generator,
):
    """
    CNVM simulation with constant-complexity update.
    """
    # pre-calculate some values
    num_agents = x.shape[0]
    max_degree_alpha = np.max(degree_alpha)
    next_event_rate = 1 / (r_imit * max_degree_alpha + r_noise) / num_agents
    noise_probability = r_noise / (r_noise + r_imit * max_degree_alpha)
    prob_factor = degree_alpha / max_degree_alpha

    # initialize
    x_traj = [np.copy(x)]
    t = 0
    t_traj = [0]

    # simulation loop
    t_store = t_delta
    while t < t_max:
        t += rng.exponential(next_event_rate)  # time of next event
        agent = rng.integers(0, num_agents)  # agent of next event
        noise = True if rng.random() < noise_probability else False

        if noise:
            new_opinion = rng.integers(0, num_opinions)
            if rng.random() < prob_noise[x[agent], new_opinion]:
                x[agent] = new_opinion
        elif rng.random() < prob_factor[agent]:
            neighbors = neighbor_list[agent]
            rand_neighbor = neighbors[rng.integers(0, len(neighbors))]
            new_opinion = x[rand_neighbor]
            if rng.random() < prob_imit[x[agent], new_opinion]:
                x[agent] = new_opinion

        if t >= t_store:
            t_store += t_delta
            x_traj.append(x.copy())
            t_traj.append(t)

    return t_traj, x_traj


@njit()
def _numba_simulate_const_complete(
    x: np.ndarray,
    t_delta: float,
    t_max: float,
    num_opinions: int,
    r_imit: float,
    r_noise: float,
    prob_imit: np.ndarray,
    prob_noise: np.ndarray,
    degree_alpha: np.ndarray,
    rng: Generator,
):
    """
    CNVM simulation for complete networks.
    """
    # pre-calculate some values
    num_agents = x.shape[0]
    max_degree_alpha = degree_alpha[0]
    next_event_rate = 1 / (r_imit * max_degree_alpha + r_noise) / num_agents
    noise_probability = r_noise / (r_noise + r_imit * max_degree_alpha)

    # initialize
    x_traj = [np.copy(x)]
    t = 0
    t_traj = [0]

    # simulation loop
    t_store = t_delta
    while t < t_max:
        t += rng.exponential(next_event_rate)  # time of next event
        agent = rng.integers(0, num_agents)  # agent of next event
        noise = True if rng.random() < noise_probability else False

        if noise:
            new_opinion = rng.integers(0, num_opinions)
            if rng.random() < prob_noise[x[agent], new_opinion]:
                x[agent] = new_opinion
        else:
            neighbor = rng.integers(0, num_agents)
            while neighbor == agent:
                neighbor = rng.integers(0, num_agents)
            new_opinion = x[neighbor]
            if rng.random() < prob_imit[x[agent], new_opinion]:
                x[agent] = new_opinion

        if t >= t_store:
            t_store += t_delta
            x_traj.append(x.copy())
            t_traj.append(t)

    return t_traj, x_traj



from typing import Protocol
import networkx as nx
import numpy as np
from numpy.random import Generator, default_rng
import time


class NetworkGenerator(Protocol):
    num_agents: int

    def __call__(self) -> nx.Graph:
        """Generate a network."""

    def __repr__(self) -> str:
        """Return string representation of the generator."""

    def abrv(self) -> str:
        """Return short description for file names."""


class ErdosRenyiGenerator:
    def __init__(
        self,
        num_agents: int,
        p: float,
        max_sample_time: float = 10,
        rng: Generator = default_rng(),
    ):
        """
        Generate Erdös-Renyi (binomial) random graphs.

        The random graph may contain isolated vertices, which is not allowed.
        In that case, the Generator samples until a valid network is found,
        or until max_sample_time seconds pass, in which case a RuntimeError is raised.

        Parameters
        ----------
        num_agents : int
        p : float
        max_sample_time : float, optional
            In seconds.
        rng : Generator, optional
            random number generator
        """
        self.num_agents = num_agents
        self.p = p
        self.max_sample_time = max_sample_time
        self.rng = rng

    def __call__(self) -> nx.Graph:
        gnp_fun = nx.erdos_renyi_graph if self.p > 0.2 else nx.fast_gnp_random_graph
        start = time.time()
        while True:
            network = gnp_fun(self.num_agents, self.p, seed=self.rng)
            if nx.number_of_isolates(network) == 0:
                return network

            if time.time() - start > self.max_sample_time:
                raise RuntimeError(
                    "Timeout. Could not generate a graph without isolated vertices."
                )

    def __repr__(self) -> str:
        return f"Erdos-Renyi random graph with p={self.p} on {self.num_agents} nodes"

    def abrv(self) -> str:
        return f"ER_p{int(self.p * 100)}_N{self.num_agents}"


class RandomRegularGenerator:
    def __init__(self, num_agents: int, d: int, rng: Generator = default_rng()):
        """
        Generate random regular graphs.

        Parameters
        ----------
        num_agents : int
        d : int
        rng : Generator, optional
            random number generator
        """
        self.num_agents = num_agents
        self.d = d
        self.rng = rng

    def __call__(self) -> nx.Graph:
        return nx.random_regular_graph(self.d, self.num_agents, seed=self.rng)

    def __repr__(self) -> str:
        return (
            f"Uniform random regular graph with d={self.d} on {self.num_agents} nodes"
        )

    def abrv(self) -> str:
        return f"regular_d{self.d}_N{self.num_agents}"


class BarabasiAlbertGenerator:
    def __init__(self, num_agents: int, m: int, rng: Generator = default_rng()):
        """
        Generate random scale-free graphs using the Barabasi-Albert model.

        Parameters
        ----------
        num_agents : int
        m : int
        rng : Generator, optional
            random number generator
        """
        self.num_agents = num_agents
        self.m = m
        self.rng = rng

    def __call__(self) -> nx.Graph:
        return nx.barabasi_albert_graph(self.num_agents, self.m, seed=self.rng)

    def __repr__(self) -> str:
        return f"Barabasi-Albert random graph on {self.num_agents} nodes"

    def abrv(self):
        return f"barabasi_m{self.m}_N{self.num_agents}"


class WattsStrogatzGenerator:
    def __init__(
        self,
        num_agents: int,
        num_neighbors: int,
        p: float,
        rng: Generator = default_rng(),
    ):
        """
        Create random small-world networks using the Watts-Strogatz model.

        Parameters
        ----------
        num_agents : int
        num_neighbors : int
        p : float
        rng : Generator, optional
            random number generator
        """
        self.num_agents = num_agents
        self.num_neighbors = num_neighbors
        self.p = p
        self.rng = rng

    def __call__(self) -> nx.Graph:
        return nx.connected_watts_strogatz_graph(
            self.num_agents, self.num_neighbors, self.p, seed=self.rng
        )

    def __repr__(self) -> str:
        return f"Watts-Strogatz random graph on {self.num_agents} nodes"

    def abrv(self):
        return f"watts_k{self.num_neighbors}_p{int(self.p * 100)}_N{self.num_agents}"


class StochasticBlockGenerator:
    def __init__(
        self,
        num_agents: int,
        p_matrix: np.ndarray,
        max_sample_time: float = 10,
        rng: Generator = default_rng(),
    ):
        """
        Creates n stochastic blocks, block i is randomly connected to block j with edge density p_matrix[i, j].

        The random graph may contain isolated vertices, which is not allowed.
        In that case, the Generator samples until a valid network is found,
        or until max_sample_time seconds pass, in which case a RuntimeError is raised.

        Parameters
        ----------
        num_agents : int
        p_matrix : np.ndarray
            (n x n) matrix of edge probabilities.
        max_sample_time : float, optional
            In seconds.
        rng : Generator, optional
            random number generator
        """
        self.p_matrix = p_matrix
        self.max_sample_time = max_sample_time

        self.num_blocks = p_matrix.shape[0]
        self.block_size = int(num_agents / self.num_blocks)
        self.num_agents = self.block_size * self.num_blocks
        self.rng = rng

    def _sample_adj_matrix(self):
        adj_matrix = np.zeros((self.num_agents, self.num_agents))
        for i in range(self.num_blocks):
            for j in range(i + 1):
                this_block = (
                    self.rng.random((self.block_size, self.block_size))
                    <= self.p_matrix[i, j]
                )
                adj_matrix[
                    i * self.block_size : (i + 1) * self.block_size,
                    j * self.block_size : (j + 1) * self.block_size,
                ] = this_block

        # only keep the lower triangle and symmetrize
        adj_matrix = np.tril(adj_matrix, -1)
        adj_matrix = adj_matrix + np.tril(adj_matrix).T
        return adj_matrix

    def __call__(self) -> nx.Graph:
        start = time.time()
        while True:
            adj_mat = self._sample_adj_matrix()
            network = nx.from_numpy_array(adj_mat)
            if nx.number_of_isolates(network) == 0:
                return network

            if time.time() - start > self.max_sample_time:
                raise RuntimeError(
                    "Timeout. Could not generate a graph without isolated vertices."
                )

    def __repr__(self) -> str:
        return f"stochastic block model with {self.num_blocks} blocks and {self.num_agents} nodes."

    def abrv(self):
        return f"sbm_b{self.num_blocks}_N{self.num_agents}"


class GridGenerator:
    def __init__(self, num_agents: int, periodic: bool = False):
        """
        Generate lattice graph in 2 dimensions.

        Parameters
        ----------
        num_agents : int
        periodic : bool, optional
        """
        self.num_agents = num_agents
        self.dim = 2  # only implemented for 2D grids for now
        self.periodic = periodic

        # find shape as square as possible
        shape0 = round(num_agents**0.5)
        while num_agents % shape0 != 0:
            shape0 += 1
        shape1 = int(num_agents / shape0)
        self.shape = shape0, shape1

    def __call__(self) -> nx.Graph:
        g = nx.grid_graph(self.shape, periodic=self.periodic)

        relabel_dict = {}
        for i, val in enumerate(g.nodes):
            relabel_dict[val] = i

        return nx.relabel_nodes(g, relabel_dict)

    def __repr__(self) -> str:
        return f"{self.shape} lattice graph graph"

    def abrv(self):
        if self.periodic:
            return f"lattice_{self.dim}D_periodic_N{self.num_agents}"
        else:
            return f"lattice_{self.dim}D_N{self.num_agents}"


class BinomialWattsStrogatzGenerator:
    def __init__(
        self,
        num_agents: int,
        num_neighbors: int,
        p_rewire: float,
        max_sample_time: float = 10,
        rng: Generator = default_rng(),
    ):
        """
        Creates a ring where each node is connected to the num_neighbors nearest neighbors.
        (num_neighbors needs to be even!)
        Then iterate through each edge and rip it out with probability p_rewire.
        Then iterate through all the possible edges that are not present and insert with such a probability,
        that in expectation the resulting graph has the same number of edges again.
        For p=1, this yields the binomial Erdös-Renyi graph G(N, K/N).

        The random graph may contain isolated vertices, which is not allowed.
        In that case, the Generator samples until a valid network is found,
        or until max_sample_time seconds pass, in which case a RuntimeError is raised.

        Parameters
        ----------
        num_agents : int
        num_neighbors : int
        p_rewire : float
        max_sample_time : float, optional
            In seconds.
        rng : Generator, optional
            random number generator
        """
        self.num_agents = num_agents
        self.num_neighbors = num_neighbors
        self.p_rewire = p_rewire
        self.max_sample_time = max_sample_time

        self.p_insert = (
            p_rewire * num_neighbors / (num_agents - 1 - (1 - p_rewire) * num_neighbors)
        )
        self.gnp_fun = (
            nx.erdos_renyi_graph if self.p_insert > 0.2 else nx.fast_gnp_random_graph
        )
        self.rng = rng

    def _sample_network(self):
        network = nx.watts_strogatz_graph(
            self.num_agents, self.num_neighbors, 0, seed=self.rng
        )

        # remove edges
        edges = np.array(network.edges)
        idx_to_keep = self.rng.random(edges.shape[0]) > self.p_rewire
        edges = edges[idx_to_keep, :]

        # insert edges
        network = self.gnp_fun(self.num_agents, self.p_insert, seed=self.rng)
        network.add_edges_from(edges)
        return network

    def __call__(self) -> nx.Graph:
        start = time.time()
        while True:
            network = self._sample_network()
            if nx.number_of_isolates(network) == 0:
                return network

            if time.time() - start > self.max_sample_time:
                raise RuntimeError(
                    "Timeout. Could not generate a graph without isolated vertices."
                )

    def __repr__(self) -> str:
        return f"Binomial Watts-Strogatz graph on {self.num_agents} nodes with p_rewire={self.p_rewire}"

    def abrv(self):
        return f"binWS_k{self.num_neighbors}_p{int(self.p_rewire * 100)}_N{self.num_agents}"
